import json
import os
import shutil
import subprocess

import click
from jupyter_client import kernelspec

VIRTUAL_ENV_VAR = 'VIRTUAL_ENV'
CONDA_PREFIX = 'CONDA_PREFIX'
CONDA_NAME = 'CONDA_DEFAULT_ENV'


@click.command()
@click.option('-n', '--name',
              help='Name of kernel.  Must provide a kernel name or run in a virtual environment.',
              default='')
@click.option('-p', '--path', type=click.Path(),
              help='Path to add to the start of PYTHONPATH.', default='')
@click.option('-l', '--list', is_flag=True, help='Get information about the current kernel')
@click.option('-d', '--delete', is_flag=True, help='Delete the existing Jupyter kernel')
def cli(name, path, list, delete):
    """Manage jupyter kernel for this virtual environment."""
    if not in_virtual_env() and not name:
        raise click.UsageError('The environment variables {} and {} are not set (usually this is '
                               'set automatically activating a conda environment or virtualenv, '
                               'respectively).  Please activate one of these and '
                               'try again!'.format(CONDA_PREFIX, VIRTUAL_ENV_VAR))
    if list:
        success, kernel, kernel_path = read_kernel(name)
        if success:
            click.secho('Found kernel "{display_name}" at {kernel_path}:'.format(
                        kernel_path=kernel_path,
                        **kernel), fg='green')
            click.secho(json.dumps(kernel, indent=2))
        else:
            click.secho('No kernel found at {}'.format(kernel_path), fg='red')
    elif delete:
        success, kernel, kernel_path = delete_kernel(name)
        if success:
            click.secho('Deleted jupyter kernel "{display_name}" from {kernel_path}:'.format(
                kernel_path=kernel_path, **kernel), fg='green')
            click.secho(json.dumps(kernel, indent=2))
        else:
            click.secho('No kernel found to delete (checked {})'.format(kernel_path), fg='red')
    else:
        success, kernel, kernel_path = install_kernel(name, click.format_filename(path))
        if success:
            click.secho(
                'Successfully installed a new jupyter kernel "{display_name}":'.format(**kernel),
                fg='green')
            click.secho(json.dumps(kernel, indent=2))
            click.secho('See {} to edit.'.format(kernel_path), fg='green')
            if need_to_install_ipykernel():
                click.secho('Remember to install ipykernel before starting a jupyter notebook!',
                            fg='red')
        else:
            click.secho('Failed to install a new jupyter kernel "{display_name}".\n'
                        'See {kernel_path} to confirm it isn\'t already there.'.format(
                            kernel_path=kernel_path, **kernel),
                        fg='red')


def in_virtual_env():
    """Check whether script is being run in a virtualenv."""
    return bool(os.getenv(VIRTUAL_ENV_VAR) or os.getenv(CONDA_PREFIX))


def delete_kernel(name):
    """Delete the jupyter kernel for the current virtualenv."""
    success, kernel, kernel_path = read_kernel(name)
    if success:
        shutil.rmtree(os.path.dirname(kernel_path))
    return success, kernel, kernel_path


def read_kernel(name):
    """Get information about the jupyter kernel for the current virtualenv."""
    display_name = get_display_name(name)
    kernel_path = get_kernel_path(display_name)
    if os.path.exists(kernel_path):
        with open(kernel_path, 'r') as buff:
            return True, json.load(buff), kernel_path
    else:
        return False, {}, kernel_path


def install_kernel(name, path):
    """Creates a jupyter kernel using the activated virtual environment."""
    display_name = get_display_name(name)
    executable = get_executable()
    env = get_env(path)
    kernel = get_kernel(display_name, executable, env)
    kernel_path = get_kernel_path(display_name)
    if confirm_kernel_path_is_safe(kernel_path):
        with open(kernel_path, 'w') as buff:
            json.dump(kernel, buff)
        return True, kernel, kernel_path
    return False, kernel, kernel_path


def get_kernel(display_name, executable, env):
    """Formats kernel in necessary format"""
    return {
        'argv': [
            executable,
            '-m',
            'ipykernel',
            '-f',
            '{connection_file}'],
        'display_name': display_name,
        'env': env,
        'language': 'python'
    }


def get_display_name(name):
    """Display name for kernel"""
    if name:
        return name
    conda_name = os.getenv(CONDA_NAME)
    if conda_name:
        return conda_name
    return os.path.basename(os.getenv(VIRTUAL_ENV_VAR))


def get_executable():
    """Get the python executable path"""
    base_path = os.getenv(CONDA_PREFIX, os.getenv(VIRTUAL_ENV_VAR))
    return os.path.join(base_path, 'bin', 'python')


def get_env(path):
    """Get any environment variables"""
    if path:
        return {'PYTHONPATH': '{}:$PYTHONPATH'.format(path)}
    else:
        return {}


def get_kernel_path(display_name):
    """Convert display name into file-safe name"""
    safe_display = "".join([s if s.isalnum() else "_" for s in display_name])
    filename = os.path.join(safe_display, 'kernel.json')
    return os.path.join(kernelspec.jupyter_data_dir(), 'kernels', filename)


def confirm_kernel_path_is_safe(kernel_path):
    """Make sure necessary directory exists and a kernel with the same name does not exist"""
    if os.path.exists(kernel_path):
        return False
    directory = os.path.dirname(kernel_path)
    if not os.path.isdir(directory):
        os.makedirs(directory)
    return True


def need_to_install_ipykernel():
    """Check if ipykernel is installed"""
    status = subprocess.call(['python', '-c', 'import ipykernel'],
                             stderr=subprocess.STDOUT,
                             stdout=open(os.devnull, 'w'),
                             close_fds=True)
    return status == 0



# -*- coding: utf-8 -*-
import os
import shutil
import tempfile
import unittest
try:
    from unittest.mock import patch  # py3
except ImportError:
    from mock import patch

from click.testing import CliRunner

from callisto import callisto


@patch('callisto.callisto.kernelspec')
class TestCallisto(unittest.TestCase):
    test_cases = (
        'easy',
        'hard/one with spaces',
        '\o/',
        '¯\\_(ツ)_/¯',
        u'\\u041c\\u0430\\u043a\\u0435\\u0434\\u043e\\u043d\\u0438\\u0458\\u0430',
    )

    @classmethod
    def setUpClass(cls):
        cls.tempdir = tempfile.mkdtemp()

    def setUp(self):
        os.environ[callisto.VIRTUAL_ENV_VAR] = self.tempdir

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tempdir)

    def tearDown(self):
        os.environ.pop(callisto.VIRTUAL_ENV_VAR)

    def test_get_kernel_path(self, mock_kernelspec):
        mock_kernelspec.jupyter_data_dir.return_value = self.tempdir
        for test_case in self.test_cases:
            filename = callisto.get_kernel_path(test_case)
            #  All filenames end with kernel.json
            self.assertTrue(filename.endswith('kernel.json'))

    @patch('callisto.callisto.get_display_name')
    def test_install_kernel(self, mock_get_display_name, mock_kernelspec):
        mock_kernelspec.jupyter_data_dir.return_value = self.tempdir
        for test_case in self.test_cases:
            mock_get_display_name.return_value = test_case
            kernel_path = callisto.get_kernel_path(test_case)
            #  Path is safe to write to
            self.assertTrue(callisto.confirm_kernel_path_is_safe(kernel_path))

            fake_env_path = 'fake_env_path'
            success, kernel, actual_kernel_path = callisto.install_kernel('', fake_env_path)

            #  Patching worked
            self.assertEqual(kernel_path, actual_kernel_path)

            # Great success
            self.assertTrue(success)

            # Added to python path
            self.assertIn(fake_env_path, kernel['env']['PYTHONPATH'])

            #  Path isn't safe any more
            self.assertFalse(callisto.confirm_kernel_path_is_safe(kernel_path))

            #  Data is sort of correct
            self.assertEqual(kernel['display_name'], test_case)

            #  No more success
            success, _, __ = callisto.install_kernel('', fake_env_path)
            self.assertFalse(success)

    def test_cli(self, mock_kernelspec):
        mock_kernelspec.jupyter_data_dir.return_value = self.tempdir
        runner = CliRunner()

        result = runner.invoke(callisto.cli)
        self.assertEqual(result.exit_code, 0)

        success, kernel, path = callisto.read_kernel('')
        self.assertTrue(success)
        self.assertIn(self.tempdir, path)
        #  Make sure tempdir is also in the python executable
        self.assertIn(self.tempdir, kernel['argv'][0])

        #  Can also confirm this with '-l'
        result = runner.invoke(callisto.cli, ["-l"])
        self.assertEqual(result.exit_code, 0)
        self.assertIn(self.tempdir, result.output)

        result = runner.invoke(callisto.cli, ["-d"])
        self.assertEqual(result.exit_code, 0)
        #  Deleting also echoes the kernel
        self.assertIn(self.tempdir, result.output)

        #  Deleting twice is handled elegantly
        result = runner.invoke(callisto.cli, ["-d"])
        self.assertEqual(result.exit_code, 0)

        result = runner.invoke(callisto.cli, ["-l"])
        self.assertEqual(result.exit_code, 0)
        #  No kernel installed
        self.assertIn('No kernel found', result.output)

        venv_name = 'pete'
        result = runner.invoke(callisto.cli, ["-n", venv_name])
        self.assertEqual(result.exit_code, 0)
        self.assertIn(self.tempdir, result.output)
        self.assertIn(venv_name, result.output)

        result = runner.invoke(callisto.cli, ["-n", venv_name])
        self.assertEqual(result.exit_code, 0)
        self.assertIn("Failed", result.output)

        # Exit without status code 0 if used outside of virtual environment
        tmp = os.environ.pop(callisto.VIRTUAL_ENV_VAR)
        result = runner.invoke(callisto.cli)
        self.assertNotEqual(result.exit_code, 0)
        os.environ[callisto.VIRTUAL_ENV_VAR] = tmp



from codecs import open
from os import path
from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as buff:
    long_description = buff.read()

setup(
    name='callisto',
    version='0.5',
    description='Create jupyter kernels from virtual environments',
    long_description=long_description,
    author='Colin Carroll',
    author_email='colcarroll@gmail.com',
    url='https://github.com/ColCarroll/callisto',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    packages=find_packages(exclude=['test']),
    install_requires=[
        'Click',
        'ipykernel'
    ],
    include_package_data=True,
    entry_points='''
    [console_scripts]
    callisto=callisto:cli
    ''',
)
